
__author__ = ["Xembie","PhaethonH","Bob Holcomb","Damien McGinnes","Robert (Tr3B) Beckebans"]
__version__ = '1.4'
__url__ = ["www.blender.org","http://xreal.sourceforge.net","http://www.katsbits.com"]

import math,time,os
from . import md3_encode

def message(log,msg):
  if log:
    log.write(msg + "\n")
  else:
    print(msg)

class md3Settings:
  def __init__(self,
               savepath,
               name,
               logtype,
               dumpall=False,
               ignoreuvs=False,
               scale=1.0,
               offsetx=0.0,
               offsety=0.0,
               offsetz=0.0):
    self.savepath = savepath
    self.name = name
    self.logtype = logtype
    self.dumpall = dumpall
    self.ignoreuvs = ignoreuvs
    self.scale = scale
    self.offsetx = offsetx
    self.offsety = offsety
    self.offsetz = offsetz

def print_md3(log,md3,dumpall):
  message(log,"Header Information")
  message(log,"Ident: " + str(md3.ident))
  message(log,"Version: " + str(md3.version))
  message(log,"Name: " + md3.name)
  message(log,"Flags: " + str(md3.flags))
  message(log,"Number of Frames: " + str(md3.numFrames))
  message(log,"Number of Tags: " + str(md3.numTags))
  message(log,"Number of Surfaces: " + str(md3.numSurfaces))
  message(log,"Number of Skins: " + str(md3.numSkins))
  message(log,"Offset Frames: " + str(md3.ofsFrames))
  message(log,"Offset Tags: " + str(md3.ofsTags))
  message(log,"Offset Surfaces: " + str(md3.ofsSurfaces))
  message(log,"Offset end: " + str(md3.ofsEnd))
  
  if dumpall:
    message(log,"Frames:")
    for f in md3.frames:
      message(log," Mins: " + str(f.mins[0]) + " " + str(f.mins[1]) + " " + str(f.mins[2]))
      message(log," Maxs: " + str(f.maxs[0]) + " " + str(f.maxs[1]) + " " + str(f.maxs[2]))
      message(log," Origin(local): " + str(f.localOrigin[0]) + " " + str(f.localOrigin[1]) + " " + str(f.localOrigin[2]))
      message(log," Radius: " + str(f.radius))
      message(log," Name: " + f.name)

    message(log,"Tags:")
    for t in md3.tags:
      message(log," Name: " + t.name)
      message(log," Origin: " + str(t.origin[0]) + " " + str(t.origin[1]) + " " + str(t.origin[2]))
      message(log," Axis[0]: " + str(t.axis[0]) + " " + str(t.axis[1]) + " " + str(t.axis[2]))
      message(log," Axis[1]: " + str(t.axis[3]) + " " + str(t.axis[4]) + " " + str(t.axis[5]))
      message(log," Axis[2]: " + str(t.axis[6]) + " " + str(t.axis[7]) + " " + str(t.axis[8]))

    message(log,"Surfaces:")
    for s in md3.surfaces:
      message(log," Ident: " + s.ident)
      message(log," Name: " + s.name)
      message(log," Flags: " + str(s.flags))
      message(log," # of Frames: " + str(s.numFrames))
      message(log," # of Shaders: " + str(s.numShaders))
      message(log," # of Verts: " + str(s.numVerts))
      message(log," # of Triangles: " + str(s.numTriangles))
      message(log," Offset Triangles: " + str(s.ofsTriangles))
      message(log," Offset UVs: " + str(s.ofsUV))
      message(log," Offset Verts: " + str(s.ofsVerts))
      message(log," Offset End: " + str(s.ofsEnd))
      message(log," Shaders:")
      for shader in s.shaders:
        message(log,"  Name: " + shader.name)
        message(log,"  Index: " + str(shader.index))
      message(log," Triangles:")
      for tri in s.triangles:
        message(log,"  Indexes: " + str(tri.indexes[0]) + " " + str(tri.indexes[1]) + " " + str(tri.indexes[2]))
      message(log," UVs:")
      for uv in s.uv:
        message(log,"  U: " + str(uv.u))
        message(log,"  V: " + str(uv.v)) 
      message(log," Verts:")
      for vert in s.verts:
        message(log,"  XYZ: " + str(vert.xyz[0]) + " " + str(vert.xyz[1]) + " " + str(vert.xyz[2]))
        message(log,"  Normal: " + str(vert.normal))

  shader_count = 0
  vert_count = 0
  tri_count = 0
  for surface in md3.surfaces:
    shader_count += surface.numShaders
    tri_count += surface.numTriangles
    vert_count += surface.numVerts
    if surface.numShaders >= md3_encode.MD3_MAX_SHADERS:
      message(log,"!Warning: Shader limit (" + str(surface.numShaders) + "/" + str(md3_encode.MD3_MAX_SHADERS) + ") reached for surface " + surface.name)
    if surface.numVerts >= md3_encode.MD3_MAX_VERTICES:
      message(log,"!Warning: Vertex limit (" + str(surface.numVerts) + "/" + str(md3_encode.MD3_MAX_VERTICES) + ") reached for surface " + surface.name)
    if surface.numTriangles >= md3_encode.MD3_MAX_TRIANGLES:
      message(log,"!Warning: Triangle limit (" + str(surface.numTriangles) + "/" + str(md3_encode.MD3_MAX_TRIANGLES) + ") reached for surface " + surface.name)
  
  if md3.numTags >= md3_encode.MD3_MAX_TAGS:
    message(log,"!Warning: Tag limit (" + str(md3.numTags) + "/" + str(md3_encode.MD3_MAX_TAGS) + ") reached for md3!")
  if md3.numSurfaces >= md3_encode.MD3_MAX_SURFACES:
    message(log,"!Warning: Surface limit (" + str(md3.numSurfaces) + "/" + str(md3_encode.MD3_MAX_SURFACES) + ") reached for md3!")
  if md3.numFrames >= md3_encode.MD3_MAX_FRAMES:
    message(log,"!Warning: Frame limit (" + str(md3.numFrames) + "/" + str(md3_encode.MD3_MAX_FRAMES) + ") reached for md3!")
    
  message(log,"Total Shaders: " + str(shader_count))
  message(log,"Total Triangles: " + str(tri_count))
  message(log,"Total Vertices: " + str(vert_count))

def save_md3(settings):
  starttime = time.clock()#start timer
  newlogpath = os.path.splitext(settings.savepath)[0] + ".log"
  if settings.logtype == "append":
    log = open(newlogpath,"a")
  elif settings.logtype == "overwrite":
    log = open(newlogpath,"w")
  else:
    log = 0
  message(log,"######################BEGIN######################")
  message(log,"Exporting selected objects...")
  ##bpy.ops.object.mode_set(mode='OBJECT')
  md3 = md3_encode.md3Object()
  md3.ident = md3_encode.MD3_IDENT
  md3.version = md3_encode.MD3_VERSION
  md3.name = settings.name
  md3.numFrames = (bpy.context.scene.frame_end + 1) - bpy.context.scene.frame_start

  # matrix to apply offset & scale - just calculate once and cache
  import mathutils
  offset_scale_mat = mathutils.Matrix.Translation((settings.offsetx, settings.offsety, settings.offsetz)) * mathutils.Matrix.Scale(settings.scale, 4)
  
  # export all selected meshes, remember empties for later...
  empties = []
  for obj in bpy.context.selected_objects:
    # ...of type mesh
    if obj.type == 'MESH':
      bpy.context.scene.frame_set(bpy.context.scene.frame_start)
      nobj = obj.to_mesh(bpy.context.scene,True,'PREVIEW')
      
      nsurface = md3_encode.md3Surface()
      nsurface.name = obj.name
      nsurface.ident = md3_encode.MD3_IDENT
      
      ignoreuvs = False
      nshader = md3_encode.md3Shader()
      #Add only 1 shader per surface/object
      try:
        #Using custom properties allows a longer string
        nshader.name = obj["md3shader"]
      except:#we should add a blank as a placeholder
        ignoreuvs = True
        nshader.name = "NULL"
      nsurface.shaders.append(nshader)
      nsurface.numShaders = 1
 
      vertlist = []

      for f,face in enumerate(nobj.faces):
        ntri = md3_encode.md3Triangle()
        if len(face.vertices) != 3:
          message(log,"Found a nontriangle face in object " + obj.name)
          continue

        for v,vert_index in enumerate(face.vertices):
          if settings.ignoreuvs or ignoreuvs:
            uv_u = 0
            uv_v = 0
          else:
            uv_u = round(obj.data.uv_textures.active.data[f].uv[v][0],5)
            uv_v = round(obj.data.uv_textures.active.data[f].uv[v][1],5)

          match = 0
          match_index = 0
          for i,vi in enumerate(vertlist):
            if vi == vert_index:
              if settings.ignoreuvs or ignoreuvs:
                match = 1
                match_index = i
              else:
                if nsurface.uv[i].u == uv_u and nsurface.uv[i].v == uv_v:
                  match = 1
                  match_index = i

          if match == 0:
            vertlist.append(vert_index)
            ntri.indexes[v] = nsurface.numVerts
            ntex = md3_encode.md3TexCoord()
            ntex.u = uv_u
            ntex.v = uv_v
            nsurface.uv.append(ntex)
            nsurface.numVerts += 1
          else:
            ntri.indexes[v] = match_index
        nsurface.triangles.append(ntri)
        nsurface.numTriangles += 1

      for frame in range(bpy.context.scene.frame_start,bpy.context.scene.frame_end + 1):
        bpy.context.scene.frame_set(frame)
        fobj = obj.to_mesh(bpy.context.scene,True,'PREVIEW')
        #fobj.calc_normals()
        # only add new frames once
        if len(md3.frames) <= frame - bpy.context.scene.frame_start:
            nframe = md3_encode.md3Frame()
            nframe.name = str(frame)
        else:
            nframe = md3.frames[frame - bpy.context.scene.frame_start]
        for vi in vertlist:
          vert = fobj.vertices[vi]
          nvert = md3_encode.md3Vert()
          nvert.normal = nvert.Encode(vert.normal)
          nvert.xyz = offset_scale_mat * obj.matrix_world * vert.co
          for i in range(0,3):
            nvert.xyz[i] = round(nvert.xyz[i], 5)
            nframe.mins[i] = min(nframe.mins[i],nvert.xyz[i])
            nframe.maxs[i] = max(nframe.maxs[i],nvert.xyz[i])
          minlength = math.sqrt(math.pow(nframe.mins[0],2) + math.pow(nframe.mins[1],2) + math.pow(nframe.mins[2],2))
          maxlength = math.sqrt(math.pow(nframe.maxs[0],2) + math.pow(nframe.maxs[1],2) + math.pow(nframe.maxs[2],2))
          nframe.radius = round(max(minlength,maxlength),5)
          nsurface.verts.append(nvert) 
        # only add if not already there
        if len(md3.frames) <= frame - bpy.context.scene.frame_start:
            md3.frames.append(nframe)
        nsurface.numFrames += 1
        bpy.data.meshes.remove(fobj)
      print(len(md3.frames), "==", md3.numFrames)
      assert(len(md3.frames) == md3.numFrames)
      md3.surfaces.append(nsurface)
      md3.numSurfaces += 1
      bpy.data.meshes.remove(nobj)

    elif obj.type == 'EMPTY':
      empties.append(obj)
  
  #write empties(=tags) in correct order - tag1frame1, tag2frame1, ..., tag1frame2, tag2frame2, ...
  md3.numTags = len(empties)
  for frame in range(bpy.context.scene.frame_start,bpy.context.scene.frame_end + 1):
    bpy.context.scene.frame_set(frame)
    for obj in empties:
      ntag = md3_encode.md3Tag()
      ntag.name = obj.name
      ntag.origin[0] = round((obj.matrix_world[0][3] * settings.scale) + settings.offsetx,5)
      ntag.origin[1] = round((obj.matrix_world[1][3] * settings.scale) + settings.offsety,5)
      ntag.origin[2] = round((obj.matrix_world[2][3] * settings.scale) + settings.offsetz,5)
      ntag.axis[0] = obj.matrix_world[0][0]
      ntag.axis[1] = obj.matrix_world[1][0]
      ntag.axis[2] = obj.matrix_world[2][0]
      ntag.axis[3] = obj.matrix_world[0][1]
      ntag.axis[4] = obj.matrix_world[1][1]
      ntag.axis[5] = obj.matrix_world[2][1]
      ntag.axis[6] = obj.matrix_world[0][2]
      ntag.axis[7] = obj.matrix_world[1][2]
      ntag.axis[8] = obj.matrix_world[2][2]
      md3.tags.append(ntag)
  
  if bpy.context.selected_objects:
    file = open(settings.savepath, "wb")
    md3.Save(file)
    print_md3(log,md3,settings.dumpall)
    file.close()
    message(log,"MD3 saved to " + settings.savepath)
    elapsedtime = round(time.clock() - starttime,5)
    message(log,"Elapsed " + str(elapsedtime) + " seconds")
  else:
    message(log,"Select an object to export!")
    
  if log:
    print("Logged to",newlogpath)
    log.close()
    
    


from bpy.props import *
import bpy

class ExportMD3(bpy.types.Operator):
  bl_idname = "export.md3"
  bl_label = "Export MD3"

  logenum = [("console","Console","log to console"),
             ("append","Append","append to log file"),
             ("overwrite","Overwrite","overwrite log file")]

  filepath = StringProperty(subtype = 'FILE_PATH',name="File Path", description="Filepath for exporting", maxlen= 1024, default= "")
  md3name = StringProperty(name="MD3 Name", description="MD3 header name / skin path (64 bytes)",maxlen=64,default="")
  md3logtype = EnumProperty(name="Save log", items=logenum, description="File logging options",default = 'console')
  md3dumpall = BoolProperty(name="Dump all", description="Dump all data for md3 to log",default=False)
  md3ignoreuvs = BoolProperty(name="Ignore UVs", description="Ignores uv influence on mesh generation. Use if uv map not made.",default=False)
  md3scale = FloatProperty(name="Scale", description="Scale all objects from world origin (0,0,0)",default=1.0,precision=5)
  md3offsetx = FloatProperty(name="Offset X", description="Transition scene along x axis",default=0.0,precision=5)
  md3offsety = FloatProperty(name="Offset Y", description="Transition scene along y axis",default=0.0,precision=5)
  md3offsetz = FloatProperty(name="Offset Z", description="Transition scene along z axis",default=0.0,precision=5)

  def execute(self, context):
    settings = md3Settings(savepath = self.properties.filepath,
                           name = self.properties.md3name,
                           logtype = self.properties.md3logtype,
                           dumpall = self.properties.md3dumpall,
                           ignoreuvs = self.properties.md3ignoreuvs,
                           scale = self.properties.md3scale,
                           offsetx = self.properties.md3offsetx,
                           offsety = self.properties.md3offsety,
                           offsetz = self.properties.md3offsetz)
    save_md3(settings)
    return {'FINISHED'}

  def invoke(self, context, event):
    wm = context.window_manager
    wm.fileselect_add(self)
    return {'RUNNING_MODAL'}